"""
Parameter / variable
"""

from __future__ import annotations
from typing import Optional, Union, TypeVar, Sequence
from contextlib import contextmanager
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf
from ._backend import global_backend as _global_backend


__all__ = ["Parameter", "set_parameter_copy_behavior_ctx"]


T = TypeVar("T")


class Parameter(Tensor[T]):
    """
    This represents a (potential trainable) parameter,
    aka ``tf.Variable`` in TensorFlow,
    wrapping to ``VariableLayer`` in RETURNN.
    """

    def __init__(
        self,
        dims_or_tensor: Union[None, Sequence[Dim], Tensor] = None,
        dtype: Optional[str] = None,
        *,
        dims: Optional[Sequence[Dim]] = None,
        sparse_dim: Optional[Dim] = None,
        trainable: Optional[bool] = None,
        auxiliary: bool = False,
        non_critical_for_restore: bool = False,
        weight_decay: Optional[float] = 0.0,
        initial: Optional[rf.init.ParamInitType] = None,
        raw_tensor: Optional[T] = None,
        device: Optional[str] = None,
    ):
        """
        :param dims_or_tensor:
        :param dims:
        :param dtype:
        :param sparse_dim:
        :param trainable: if True, and optimizer would do updates to this parameter in training mode
        :param auxiliary: if True, this indicates that this parameter should not be transformed by transformations
          such as weight normalization. One example are running statistics, as used for batch normalization.
          This usually implies that the parameter is not trainable, i.e. not to be updated by the optimizer,
          but usually has some custom update.
          This flag is not passed on to RETURNN but just used here for returnn-common logic.
        :param non_critical_for_restore: if True, this parameter is not critical for restoring a model.
        :param weight_decay:
        :param initial:
        :param raw_tensor:
        :param device:
        """
        if dims_or_tensor is None:
            if dims is None:
                raise ValueError("rf.Parameter: dims must be set.")
        elif isinstance(dims_or_tensor, (tuple, list)):
            if dims is not None:
                raise ValueError("rf.Parameter: dims is set twice.")
            dims = dims_or_tensor
        elif isinstance(dims_or_tensor, Tensor):
            if initial is not None:
                raise ValueError("rf.Parameter: initial is set twice.")
            initial = dims_or_tensor
            if dims is None:
                dims = initial.dims
            if dtype is None:
                dtype = initial.dtype
            if sparse_dim is None:
                sparse_dim = initial.sparse_dim
        else:
            raise TypeError(f"rf.Parameter: invalid type for dims_or_tensor: {type(dims_or_tensor)}")
        if not all(isinstance(dim, Dim) for dim in dims):
            raise TypeError(f"rf.Parameter: shape {dims} must be a sequence of Dim")
        if len(dims) != len(set((d, d.match_priority) for d in dims)):
            raise ValueError(f"rf.Parameter: shape {dims} dims must be unique")
        super(Parameter, self).__init__(
            "parameter",
            dims=dims,
            dtype=dtype or (rf.get_default_float_dtype() if not sparse_dim else rf.get_default_array_index_dtype()),
            sparse_dim=sparse_dim,
        )
        if raw_tensor is not None:
            self.raw_tensor = raw_tensor
        else:
            self.raw_tensor = _global_backend.create_parameter_raw(self, device=device)
        self._trainable = None  # type: Optional[bool]
        self._auxiliary = auxiliary
        self._non_critical_for_restore = non_critical_for_restore
        self._weight_decay = weight_decay
        self._initial = None  # type: Optional[rf.init.ParamInitType]

        self.trainable = trainable  # use setter
        self.initial = initial  # use setter

    def __copy__(self):
        if _copy_behavior == "share":
            return self
        # Should return new copy. https://github.com/rwth-i6/returnn_common/pull/215#issuecomment-1269651064
        # Note that the values are *not* copied, but rather it will use the same param init scheme.
        assert _copy_behavior == "copy_init"
        res = type(self)(
            dims=self.dims,
            dtype=self.dtype,
            trainable=self.trainable,
            auxiliary=self.auxiliary,
            non_critical_for_restore=self.non_critical_for_restore,
            weight_decay=self.weight_decay,
        )
        res.initial = self.initial
        return res

    def __deepcopy__(self, memo=None):
        if _copy_behavior == "share":
            return self

        # Should return new copy. https://github.com/rwth-i6/returnn_common/pull/215#issuecomment-1269651064
        # Note that the values are *not* copied, but rather it will use the same param init scheme.
        from copy import deepcopy

        assert _copy_behavior == "copy_init"
        res = self.__copy__()
        if isinstance(self.initial, rf.init.ParamInit):
            res.initial = deepcopy(self.initial, memo=memo)  # noqa
        else:
            res.initial = self.initial
        return res

    @property
    def initial(self) -> Optional[rf.init.ParamInitType]:
        """initial value of the parameter"""
        return self._initial

    @initial.setter
    def initial(self, value: Optional[rf.init.ParamInitType]):
        # Keep the original ParamInit, so that copies of the Parameter would have a different initial random value.
        # https://github.com/rwth-i6/returnn_common/issues/216
        self._initial = value

        if isinstance(value, rf.init.ParamInit):
            value = value(dims=self.dims, dtype=self.dtype, device=self.device)
        self._raw_backend.set_parameter_initial_value(self, value)

    def assign(self, value: Union[Tensor, rf.RawTensorTypes]):
        """
        Assign new value to this parameter.
        This will also update the allocated raw tensor inplace.

        For graph-based backends, handling the control flow is up to the backend,
        e.g.~making sure it is being executed in the right order,
        in the right control flow context, and at all.
        There is no op or anything like that returned here which the user needs to take care of.
        So the user can think of it just as imperative eager-style code.
        """
        self._raw_backend.parameter_assign(
            self, rf.convert_to_tensor(value, _backend=self._raw_backend, device=self.device), op="assign"
        )

    def assign_add(self, value: Union[Tensor, rf.RawTensorTypes]):
        """
        Add value to this parameter.
        This will also update the raw tensor.
        See :func:`assign`.
        """
        self._raw_backend.parameter_assign(
            self, rf.convert_to_tensor(value, _backend=self._raw_backend, device=self.device), op="add"
        )

    def assign_key(
        self,
        axis: Union[Dim, Sequence[Dim]],
        key: rf.ItemKeyType,
        key_dim: Optional[Union[Dim, Sequence[Optional[Dim]]]],
        value: Union[Tensor, rf.RawTensorTypes],
    ):
        """
        Basically var[key] = value, if axis is the first axis, or otherwise accordingly.
        Note that the __setitem__ API is not supported because it depends on the order of axes,
        but this here is the equivalent function.
        See :func:`assign`.
        """
        self._raw_backend.parameter_assign_key(
            self,
            axis=axis,
            key=key,
            key_dim=key_dim,
            value=rf.convert_to_tensor(value, _backend=self._raw_backend, device=self.device),
            op="assign",
        )

    def to(self, *, device: Optional[str] = None, dtype: Optional[str] = None):
        """
        Move the parameter to the specified device, and/or the specified dtype.

        Note: This is an inplace operation.
        raw_tensor might point to a new raw tensor (or parameter) afterward.
        """
        self._raw_backend.parameter_move_to(self, device=device, dtype=dtype)

    @property
    def weight_decay(self) -> float:
        """
        Weight decay, which is equivalent to L2 loss on the parameters for SGD.
        On RETURNN side, whether this is handled separately or is part of the main loss,
        can be controlled via the ``decouple_constraints`` config option.
        https://github.com/rwth-i6/returnn_common/issues/59#issuecomment-1073913421
        """
        return self._weight_decay or 0.0

    @weight_decay.setter
    def weight_decay(self, value: Optional[float]):
        self._weight_decay = value

    @property
    def trainable(self) -> Optional[bool]:
        """trainable"""
        return self._trainable

    @trainable.setter
    def trainable(self, trainable: Optional[bool]):
        self._trainable = trainable
        if trainable is None:
            if self.auxiliary:
                trainable = False
            elif self.dtype.startswith("int"):
                trainable = False
            else:
                trainable = True
        self._raw_backend.set_parameter_trainable(self, trainable)

    @property
    def auxiliary(self) -> bool:
        """auxiliary"""
        return self._auxiliary

    @auxiliary.setter
    def auxiliary(self, value: bool):
        self._auxiliary = value

    @property
    def non_critical_for_restore(self) -> bool:
        """non_critical_for_restore"""
        return self._non_critical_for_restore

    @non_critical_for_restore.setter
    def non_critical_for_restore(self, value: bool):
        self._non_critical_for_restore = value


_copy_behavior = "copy_init"


@contextmanager
def set_parameter_copy_behavior_ctx(mode: str):
    """
    Set the copy behavior for parameters.

    :param mode: "copy_init" or "share".
        "copy_init" (default when not set): When copying a parameter, a new parameter will be created.
            It is not a shared parameter.
            It does not copy the values, but it just copies the param init scheme.
        "share": The parameter will not be copied, it will just return itself.
    """
    global _copy_behavior
    assert mode in ("copy_init", "share")
    old = _copy_behavior
    try:
        _copy_behavior = mode
        yield
    finally:
        _copy_behavior = old



"""
RETURNN frontend (returnn.frontend) tests
"""

from __future__ import annotations
import _setup_test_env  # noqa
from typing import Sequence
from returnn.tensor import Tensor, Dim, TensorDict, batch_dim
from rf_utils import run_model


def _setup_lovely_tensors():
    try:
        import lovely_tensors
    except ImportError:
        return
    lovely_tensors.monkey_patch()


_setup_lovely_tensors()


def test_conformer():
    # This test needs a huge stack size currently, due to the way RETURNN layer construction works currently.
    # On RETURNN side, there is the option flat_net_construction to solve this,
    # however, it's experimental and also does not work for this case.
    # https://github.com/rwth-i6/returnn/issues/957
    # https://stackoverflow.com/a/16248113/133374
    import resource
    import sys

    try:
        resource.setrlimit(resource.RLIMIT_STACK, (2**29, -1))
    except Exception as exc:
        print(f"resource.setrlimit {type(exc).__name__}: {exc}")
    sys.setrecursionlimit(10**6)

    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    from returnn.frontend.encoder.conformer import ConformerEncoder, ConformerConvSubsample

    # noinspection PyShadowingNames
    def _forward_step(*, model: ConformerEncoder, extern_data: TensorDict):
        out, out_spatial_dim = model(extern_data["data"], in_spatial_dim=time_dim)
        out.mark_as_default_output(shape=(batch_dim, out_spatial_dim, model.out_dim))

    run_model(
        extern_data,
        lambda *, epoch, step: ConformerEncoder(
            in_dim,
            Dim(14, name="out"),
            ff_dim=Dim(17, name="ff"),
            input_layer=ConformerConvSubsample(
                in_dim,
                out_dims=[Dim(32, name="conv1"), Dim(64, name="conv2")],
                filter_sizes=[(3, 3), (3, 3)],
                pool_sizes=[(2, 1), (2, 1)],
            ),
            num_heads=2,
            num_layers=2,
        ),
        _forward_step,
    )


def test_e_branchformer():
    # https://arxiv.org/abs/2210.00077
    # https://github.com/espnet/espnet/tree/master/egs2/librispeech/asr1#e-branchformer
    # https://github.com/espnet/espnet/blob/master/egs2/librispeech/asr1/conf/tuning/train_asr_e_branchformer.yaml

    import numpy as np
    import torch
    import returnn.frontend as rf
    from returnn.util.debug import PyTracer, check_py_traces_rf_to_pt_equal

    rf.select_backend_torch()
    rf.set_random_seed(42)

    from returnn.datasets.util.vocabulary import Vocabulary

    vocab = Vocabulary.create_vocab_from_labels(
        ["<blank>", "<BOS>", "<EOS>"] + [f"label-{i}" for i in range(17)], eos_label="<EOS>", bos_label="<BOS>"
    )
    target_dim = Dim(vocab.num_labels, name="vocab", vocab=vocab)

    import os
    import tempfile
    import textwrap
    import atexit

    from espnet2.tasks.asr import ASRTask
    from espnet2.asr.espnet_model import ESPnetASRModel
    from espnet2.asr.encoder.e_branchformer_encoder import EBranchformerEncoder, EBranchformerEncoderLayer
    from espnet.nets.pytorch_backend.transformer.attention import RelPositionMultiHeadedAttention
    from espnet.nets.pytorch_backend.transformer.embedding import RelPositionalEncoding
    from espnet2.asr.layers.cgmlp import ConvolutionalGatingMLP, ConvolutionalSpatialGatingUnit

    # Code copied and adapted from i6_experiments/users/zeyer/experiments/exp2023_04_25_rf/espnet.py.
    espnet_config_file = tempfile.mktemp(suffix=".yaml", prefix="espnet_config_")
    atexit.register(os.remove, espnet_config_file)
    with open(espnet_config_file, "w") as f:
        f.write(
            textwrap.dedent(
                # Copied from ESPnet egs2/librispeech/asr1/conf/tuning/train_asr_e_branchformer.yaml.
                # Then reduced dimensions/size for the test case, and slightly simplified.
                # language=yaml
                """\
                encoder: e_branchformer
                encoder_conf:
                    output_size: 50
                    attention_heads: 2
                    attention_layer_type: rel_selfattn
                    pos_enc_layer_type: rel_pos
                    rel_pos_type: latest
                    cgmlp_linear_units: 310
                    cgmlp_conv_kernel: 3
                    use_linear_after_conv: false
                    gate_activation: identity
                    num_blocks: 2
                    dropout_rate: 0.0
                    positional_dropout_rate: 0.0
                    attention_dropout_rate: 0.0
                    input_layer: conv2d
                    layer_drop_rate: 0.0
                    linear_units: 100
                    positionwise_layer_type: linear
                    macaron_ffn: true
                    use_ffn: true
                    merge_conv_kernel: 3

                decoder: transformer
                decoder_conf:
                    attention_heads: 2
                    linear_units: 200
                    num_blocks: 1
                    dropout_rate: 0.0
                    positional_dropout_rate: 0.0
                    self_attention_dropout_rate: 0.0
                    src_attention_dropout_rate: 0.0
                    layer_drop_rate: 0.0

                model_conf:
                    ctc_weight: 0.3
                    lsm_weight: 0.1
                    length_normalized_loss: false

                frontend_conf:
                    n_fft: 512
                    hop_length: 160
                """
            )
        )

    parser = ASRTask.get_parser()
    args = parser.parse_args(["--config", espnet_config_file])
    args.token_list = target_dim.vocab.labels
    args.model_conf["sym_sos"] = target_dim.vocab.labels[vocab.bos_label_id]
    args.model_conf["sym_eos"] = target_dim.vocab.labels[vocab.eos_label_id]

    model = ASRTask.build_model(args)
    assert isinstance(model, ESPnetASRModel)
    print("Target dim:", target_dim)
    print("Vocab size:", model.vocab_size)
    print("Vocab:", vocab.labels[:5], "...", vocab.labels[-5:])
    print("Ignore:", model.ignore_id)
    print("Blank:", model.blank_id)
    print("SOS/EOS:", model.sos, model.eos)

    _torch_default_device = "cpu"
    rnd = np.random.RandomState(42)
    batch_dim_ = Dim(5, name="batch")
    raw_audio_spatial_dim = Dim(
        rf.convert_to_tensor(torch.tensor([8_457, 10_237, 17_145, 5_141, 12_015], device="cpu"), dims=[batch_dim_]),
        name="audio",
    )
    targets_spatial_dim = Dim(
        rf.convert_to_tensor(torch.tensor([11, 10, 8, 7, 5], device="cpu"), dims=[batch_dim_]), name="dec"
    )
    raw_audio = rf.convert_to_tensor(
        torch.tensor(
            rnd.randn(batch_dim_.dimension, raw_audio_spatial_dim.dyn_size_ext.raw_tensor.max()).astype(np.float32),
            device=_torch_default_device,
        ),
        dims=[batch_dim_, raw_audio_spatial_dim],
    )
    targets = rf.convert_to_tensor(
        torch.tensor(
            rnd.randint(
                3, target_dim.dimension, (batch_dim_.dimension, targets_spatial_dim.dyn_size_ext.raw_tensor.max())
            ),
            device=_torch_default_device,
        ),
        dims=[batch_dim_, targets_spatial_dim],
    )

    # espnet2.asr.espnet_model.ESPnetASRModel.forward
    #   espnet2.asr.espnet_model.ESPnetASRModel.encode
    #     espnet2.asr.frontend.default.DefaultFrontend.forward
    #       espnet2.asr.frontend.default.DefaultFrontend._compute_stft
    #       (espnet.nets.pytorch_backend.frontends.frontend.Frontend.forward) (does nothing)
    #       espnet2.layers.log_mel.LogMel.forward
    #     espnet2.layers.utterance_mvn.UtteranceMVN.forward
    #     (No preencoder)
    #     espnet2.asr.encoder.e_branchformer_encoder.EBranchformerEncoder.forward
    #       espnet.nets.pytorch_backend.transformer.subsampling.Conv2dSubsampling.forward
    #         espnet.nets.pytorch_backend.transformer.embedding.RelPositionalEncoding
    #       espnet2.asr.encoder.e_branchformer_encoder.EBranchformerEncoderLayer.forward
    #         espnet.nets.pytorch_backend.transformer.positionwise_feed_forward.PositionwiseFeedForward.forward
    #         espnet.nets.pytorch_backend.transformer.attention.RelPositionMultiHeadedAttention.forward
    #         espnet2.asr.layers.cgmlp.ConvolutionalGatingMLP.forward
    #         espnet.nets.pytorch_backend.transformer.positionwise_feed_forward.PositionwiseFeedForward.forward
    #   (CTC loss, Att-decoder)
    with PyTracer(
        [
            ESPnetASRModel.forward,
            ESPnetASRModel.encode,
            EBranchformerEncoder.forward,
            EBranchformerEncoderLayer.forward,
            RelPositionMultiHeadedAttention.forward,
            RelPositionalEncoding.forward,
            ConvolutionalGatingMLP.forward,
            ConvolutionalSpatialGatingUnit.forward,
        ],
        torch.Tensor,
    ) as trace_espnet, torch.no_grad():
        loss, stats, weight = model(
            speech=raw_audio.raw_tensor,
            speech_lengths=raw_audio_spatial_dim.dyn_size,
            text=targets.raw_tensor.to(torch.int64),  # targets are without EOS; added internally via add_sos_eos
            text_lengths=targets_spatial_dim.dyn_size,
        )

    # ESPnet usually does divide the loss by num seqs (batch dim) but not by seq length.
    print("ESPnet model loss:", loss)
    print(" stats:", stats)

    # I think we can skip the log-mel feat extraction and the conv frontend for now
    # (although it would be interesting to do that later as well)
    # and only cover the main E-Branchformer encoder itself
    # (also skipping CTC and the att-dec on top).

    # features_raw = trace_espnet.captured_locals[ESPnetASRModel.encode][0]["feats"][-1]  # [B,T',D=80]
    enc_in_raw = trace_espnet.captured_locals[EBranchformerEncoderLayer.forward][0]["x"][0]  # [B,T,D=50]
    mask_raw = trace_espnet.captured_locals[EBranchformerEncoderLayer.forward][0]["mask"][0].squeeze(dim=1)  # [B,T]
    assert mask_raw[:, 0].all()  # sanity check
    enc_out_raw = trace_espnet.captured_locals[ESPnetASRModel.forward][0]["encoder_out"][-1]  # [B,T,D=50]

    first_layer = model.encoder.encoders[0]
    assert isinstance(first_layer, EBranchformerEncoderLayer)

    enc_seq_lens_raw = torch.sum(mask_raw, dim=1).to(torch.int32)  # [B]
    batch_dim.dyn_size_ext = rf.convert_to_tensor(torch.tensor(enc_seq_lens_raw.shape[0]))
    enc_seq_lens = Tensor("enc_seq_lens", [batch_dim], dtype="int32", raw_tensor=enc_seq_lens_raw)
    enc_spatial_dim = Dim(enc_seq_lens, name="enc")
    model_dim = Dim(first_layer.size, name="model")
    enc_in = Tensor("enc_in", [batch_dim, enc_spatial_dim, model_dim], dtype="float32", raw_tensor=enc_in_raw)

    from returnn.frontend.encoder.conformer import ConformerEncoder
    from returnn.frontend.encoder.e_branchformer import EBranchformerLayer, FeedForwardConvGated

    model_rf = ConformerEncoder(
        in_dim=model_dim,
        out_dim=model_dim,
        input_layer=None,
        num_layers=len(model.encoder.encoders),
        encoder_layer=rf.build_dict(
            EBranchformerLayer,
            ff_dim=first_layer.feed_forward.w_1.out_features,
            num_heads=first_layer.attn.h,
            cgmlp_ff_dim=first_layer.cgmlp.channel_proj2.in_features,  # half the cgmlp_linear_units
            cgmlp_conv_kernel=first_layer.cgmlp.csgu.conv.kernel_size[0],
            merge_conv_kernel=first_layer.depthwise_conv_fusion.kernel_size[0],
        ),
        input_dropout=0.0,
        dropout=0.0,
        att_dropout=0.0,
    )

    from returnn.frontend.conversions.espnet_e_branchformer import import_params_espnet_e_branchformer_layer_to_rf

    assert len(model.encoder.encoders) == len(model_rf.layers)
    for layer, layer_rf in zip(model.encoder.encoders, model_rf.layers):
        assert isinstance(layer, EBranchformerEncoderLayer)
        assert isinstance(layer_rf, EBranchformerLayer)
        import_params_espnet_e_branchformer_layer_to_rf(layer, layer_rf)

    with PyTracer(
        [
            ConformerEncoder.__call__,
            EBranchformerLayer.__call__,
            rf.RelPosSelfAttention.__call__,
            rf.relative_positional_encoding,
            FeedForwardConvGated.__call__,
        ],
        Tensor,
    ) as trace_rf, torch.no_grad():
        enc_out, _ = model_rf(enc_in, in_spatial_dim=enc_spatial_dim)
    enc_out = enc_out.copy_transpose((batch_dim, enc_spatial_dim, model_dim))
    enc_out = enc_out.copy_masked(0.0)

    first_rf_layer: EBranchformerLayer = model_rf.layers[0]
    num_heads_dim: Dim = first_rf_layer.self_att.num_heads
    key_dim_per_head: Dim = first_rf_layer.self_att.key_dim_per_head
    cgmlp_ff_dim = first_rf_layer.cgmlp.ff_dim

    def _tensor(x: torch.Tensor, name: str, dims: Sequence[Dim]) -> Tensor:
        return rf.convert_to_tensor(x, name=name, dims=dims)

    from returnn.frontend.conversions.espnet_e_branchformer import _reorder_rel_pos_emb_espnet_to_rf

    check_py_traces_rf_to_pt_equal(
        trace_rf.captured_locals,
        trace_espnet.captured_locals,
        [
            (
                (EBranchformerLayer.__call__, 0, "x_ffn1_ln", 0),
                (EBranchformerEncoderLayer.forward, 0, "x", 1),
                (batch_dim, enc_spatial_dim, model_dim),
            ),
            (
                (EBranchformerLayer.__call__, 0, "x_ffn1_out", 0),
                (EBranchformerEncoderLayer.forward, 0, "x", 2),
                (batch_dim, enc_spatial_dim, model_dim),
            ),
            (
                (EBranchformerLayer.__call__, 0, "x_mhsa_ln", 0),
                (EBranchformerEncoderLayer.forward, 0, "x1", 1),
                (batch_dim, enc_spatial_dim, model_dim),
            ),
            (
                (rf.RelPosSelfAttention.__call__, 0, "source", 0),
                (RelPositionMultiHeadedAttention.forward, 0, "query", 0),
                (batch_dim, enc_spatial_dim, model_dim),
            ),
            # Also see test_rope_causal_self_att for similar checks.
            (
                (rf.RelPosSelfAttention.__call__, 0, "q", 0),
                (RelPositionMultiHeadedAttention.forward, 0, "q", 0),
                lambda x, *, name, **_: _tensor(
                    x.transpose(1, 2), name, [batch_dim, enc_spatial_dim, num_heads_dim, key_dim_per_head]
                ),
            ),
            (
                (rf.RelPosSelfAttention.__call__, 0, "v", 0),
                (RelPositionMultiHeadedAttention.forward, 0, "v", 0),
                lambda x, *, name, **_: _tensor(
                    x.transpose(1, 2), name, [batch_dim, enc_spatial_dim, num_heads_dim, key_dim_per_head]
                ),
            ),
            (
                (rf.RelPosSelfAttention.__call__, 0, "q_with_bias_u", 0),
                (RelPositionMultiHeadedAttention.forward, 0, "q_with_bias_u", 0),
                (batch_dim, num_heads_dim, enc_spatial_dim, key_dim_per_head),
            ),
            # Check RelPositionalEncoding vs our relative_positional_encoding
            # Currently disabled this check, as the dim tags are different now...
            # (
            #     (rf.RelPosSelfAttention.__call__, 0, "pos_emb", 0),
            #     (RelPositionMultiHeadedAttention.forward, 0, "pos_emb", 0),
            #     lambda x, **_: _tensor(
            #         _reorder_rel_pos_emb_espnet_to_rf(x.squeeze(dim=0)),
            #         "pos_emb",
            #         [enc_spatial_dim - 1 + enc_spatial_dim, model_dim],
            #     ),
            # ),
            (
                (EBranchformerLayer.__call__, 0, "x_mhsa", 0),
                (EBranchformerEncoderLayer.forward, 0, "x_att", 0),
                (batch_dim, enc_spatial_dim, model_dim),
            ),
            (
                (FeedForwardConvGated.__call__, 0, "x_g", 0),
                (ConvolutionalSpatialGatingUnit.forward, 0, "x_g", 0),
                (batch_dim, enc_spatial_dim, cgmlp_ff_dim),
            ),
            (
                (EBranchformerLayer.__call__, 0, "x_cgmlp", 0),
                (EBranchformerEncoderLayer.forward, 0, "x2", 2),
                (batch_dim, enc_spatial_dim, model_dim),
            ),
            (
                (ConformerEncoder.__call__, 0, "x", -1),
                (ESPnetASRModel.forward, 0, "encoder_out", -1),
                (batch_dim, enc_spatial_dim, model_dim),
            ),
        ],
    )

    # Final check.
    # Actually the check_py_traces_rf_to_pt_equal should already have covered also this final output,
    # but anyway do it again now to be sure.
    assert enc_out_raw.shape == enc_out.raw_tensor.shape  # [B,T,D]
    assert enc_out_raw.shape[:1] == enc_seq_lens_raw.shape  # [B]
    for b in range(enc_out_raw.shape[0]):
        seq_len = enc_seq_lens_raw[b]
        torch.testing.assert_allclose(enc_out_raw[b, :seq_len], enc_out.raw_tensor[b, :seq_len], rtol=1e-5, atol=1e-5)
    # Check that there is sth non-zero.
    assert enc_seq_lens_raw.max() > 0
    assert torch.mean(enc_out.raw_tensor**2) > 0.1
    print("All matching!")



"""
Convolution, transposed convolution, pooling
"""

from __future__ import annotations
from typing import Optional, Sequence, Tuple, Union
from returnn.util.basic import next_type_attrib_in_mro_chain
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf


__all__ = [
    "conv",
    "Conv1d",
    "Conv2d",
    "Conv3d",
    "transposed_conv",
    "TransposedConv1d",
    "TransposedConv2d",
    "TransposedConv3d",
    "pool",
    "max_pool",
    "max_pool1d",
    "pool1d",
    "pool2d",
    "pool3d",
    "make_conv_out_spatial_dims",
]


# noinspection PyAbstractClass
class _ConvOrTransposedConv(rf.Module):
    """
    Base class for both convolution and transposed convolution.
    """

    nd: Optional[int] = None
    _transposed: bool
    groups: Optional[int] = None

    def __init__(
        self,
        in_dim: Dim,
        out_dim: Dim,
        filter_size: Union[Sequence[Union[int, Dim]], int, Dim],
        *,
        padding: str,
        with_bias: bool,
    ):
        """
        :param Dim in_dim:
        :param Dim out_dim:
        :param filter_size: (width,), (height,width) or (depth,height,width) for 1D/2D/3D conv.
          the input data ndim must match, or you can add dimensions via input_expand_dims or input_add_feature_dim.
          it will automatically swap the batch-dim to the first axis of the input data.
        :param padding: "same" or "valid"
        :param with_bias:
        """
        super().__init__()
        assert isinstance(in_dim, Dim) and isinstance(out_dim, Dim)
        self.in_dim = rf.dim_match_priority_when_needed(in_dim, out_dim)
        self.out_dim = out_dim
        if isinstance(filter_size, (int, Dim)):
            if self.nd in (None, 1):
                filter_size = [filter_size]
            else:
                filter_size = [filter_size] * self.nd
        assert isinstance(filter_size, (tuple, list))
        if self.nd:
            assert self.nd == len(filter_size)
        else:
            self.nd = len(filter_size)
        self.filter_size = [
            s if isinstance(s, Dim) else Dim(s, name=f"filter-dim{i}") for i, s in enumerate(filter_size)
        ]
        self.padding = padding
        filter_in_dim = in_dim
        if self.groups is not None and self.groups > 1:
            assert not self._transposed  # not implemented
            filter_in_dim //= self.groups
        filter_in_dim = rf.dim_match_priority_when_needed(filter_in_dim, self.out_dim)
        self.filter_in_dim = filter_in_dim
        self.filter = rf.Parameter(
            ([self.out_dim, self.filter_in_dim] if not self._transposed else [self.in_dim, self.out_dim])
            + self.filter_size
        )
        self.filter.initial = rf.init.Glorot()
        self.with_bias = with_bias
        self.bias = None  # type: Optional[rf.Parameter]
        if self.with_bias:
            self.bias = rf.Parameter([self.out_dim])
            self.bias.initial = 0.0

    def _call_nd1(
        self,
        source: Tensor,
        *,
        in_spatial_dim: Dim,
        out_spatial_dim: Optional[Dim] = None,
    ) -> Tuple[Tensor, Dim]:
        assert self.nd == 1
        out, (out_spatial_dim,) = next_type_attrib_in_mro_chain(self.__class__, "__call__", self.__class__._call_nd1)(
            self,
            source,
            in_spatial_dims=[in_spatial_dim],
            out_spatial_dims=[out_spatial_dim] if out_spatial_dim else None,
        )
        return out, out_spatial_dim


class _Conv(_ConvOrTransposedConv):
    """
    A generic convolution layer which supports 1D, 2D and 3D convolution.
    Base class for :class:`Conv1d`, :class:`Conv2d`, :class:`Conv3d`.
    """

    _transposed = False

    # noinspection PyShadowingBuiltins,PyShadowingNames
    def __init__(
        self,
        in_dim: Dim,
        out_dim: Dim,
        filter_size: Union[Sequence[Union[int, Dim]], int, Dim],
        *,
        padding: str,
        strides: Optional[Union[int, Sequence[int]]] = None,
        dilation_rate: Optional[Union[int, Sequence[int]]] = None,
        groups: Optional[int] = None,
        with_bias: bool = True,
    ):
        """
        :param Dim in_dim:
        :param Dim out_dim:
        :param filter_size: (width,), (height,width) or (depth,height,width) for 1D/2D/3D conv.
          the input data ndim must match, or you can add dimensions via input_expand_dims or input_add_feature_dim.
          it will automatically swap the batch-dim to the first axis of the input data.
        :param str padding: "same" or "valid"
        :param int|Sequence[int] strides: strides for the spatial dims,
          i.e. length of this tuple should be the same as filter_size, or a single int.
        :param int|Sequence[int] dilation_rate: dilation for the spatial dims
        :param int groups: grouped convolution
        :param bool with_bias: if True, will add a bias to the output features
        """
        self.groups = groups
        super().__init__(in_dim=in_dim, out_dim=out_dim, filter_size=filter_size, padding=padding, with_bias=with_bias)
        if isinstance(strides, int):
            strides = [strides] * self.nd
        self.strides = strides
        self.dilation_rate = dilation_rate

    def __call__(
        self,
        source: Tensor,
        *,
        in_spatial_dims: Sequence[Dim],
        out_spatial_dims: Optional[Sequence[Dim]] = None,
    ) -> Tuple[Tensor, Sequence[Dim]]:
        return conv(
            source,
            in_dim=self.in_dim,
            out_dim=self.out_dim,
            in_spatial_dims=in_spatial_dims,
            out_spatial_dims=out_spatial_dims,
            filter=self.filter,
            filter_size=self.filter_size,
            padding=self.padding,
            strides=self.strides,
            dilation_rate=self.dilation_rate,
            groups=self.groups,
            bias=self.bias if self.with_bias else None,
        )


# noinspection PyShadowingBuiltins
def conv(
    source: Tensor,
    *,
    in_dim: Dim,
    out_dim: Dim,
    in_spatial_dims: Sequence[Dim],
    out_spatial_dims: Optional[Sequence[Dim]] = None,
    filter: Tensor,
    filter_size: Sequence[Dim],  # to have the order well-defined
    padding: str,
    strides: Optional[Union[int, Sequence[int]]] = None,
    dilation_rate: Optional[Union[int, Sequence[int]]] = None,
    groups: Optional[int] = None,
    bias: Optional[Tensor] = None,
) -> Tuple[Tensor, Sequence[Dim]]:
    """convolution"""
    for in_spatial_dim in in_spatial_dims:
        if in_spatial_dim not in source.dims:
            raise ValueError(f"conv: source {source} does not have spatial dim {in_spatial_dim}")
    # noinspection PyProtectedMember
    out, out_spatial_dims = source._raw_backend.conv(
        source,
        in_dim=in_dim,
        out_dim=out_dim,
        in_spatial_dims=in_spatial_dims,
        out_spatial_dims=out_spatial_dims,
        filter=filter,
        filter_size=filter_size,
        padding=padding,
        strides=strides,
        dilation_rate=dilation_rate,
        groups=groups,
        bias=bias,
    )
    return out, out_spatial_dims


class Conv1d(_Conv):
    """
    1D convolution
    """

    nd = 1

    def __init__(
        self,
        in_dim: Dim,
        out_dim: Dim,
        filter_size: Union[int, Dim],
        *,
        padding: str,
        strides: Optional[int] = None,
        dilation_rate: Optional[int] = None,
        groups: Optional[int] = None,
        with_bias: bool = True,
    ):
        """
        :param Dim in_dim:
        :param Dim out_dim:
        :param int|Dim filter_size:
        :param str padding: "same" or "valid"
        :param int|None strides: strides for the spatial dims,
          i.e. length of this tuple should be the same as filter_size, or a single int.
        :param int|None dilation_rate: dilation for the spatial dims
        :param int groups: grouped convolution
        :param bool with_bias: if True, will add a bias to the output features
        """
        super().__init__(
            in_dim=in_dim,
            out_dim=out_dim,
            filter_size=[filter_size],
            padding=padding,
            strides=strides,
            dilation_rate=dilation_rate,
            groups=groups,
            with_bias=with_bias,
        )

    __call__ = _ConvOrTransposedConv._call_nd1


class Conv2d(_Conv):
    """
    2D convolution
    """

    nd = 2


class Conv3d(_Conv):
    """
    3D convolution
    """

    nd = 3


class _TransposedConv(_ConvOrTransposedConv):
    """
    Transposed convolution, sometimes also called deconvolution.
    See :func:`tf.nn.conv2d_transpose` (currently we support 1D/2D).
    """

    nd: Optional[int] = None
    _transposed = True

    # noinspection PyShadowingBuiltins,PyShadowingNames
    def __init__(
        self,
        in_dim: Dim,
        out_dim: Dim,
        filter_size: Sequence[Union[int, Dim]],
        *,
        padding: str,
        remove_padding: Union[Sequence[int], int] = 0,
        output_padding: Optional[Union[Sequence[Optional[int]], int]] = None,
        strides: Optional[Sequence[int]] = None,
        with_bias: bool = True,
    ):
        """
        :param Dim in_dim:
        :param Dim out_dim:
        :param list[int] filter_size:
        :param list[int]|None strides: specifies the upscaling. by default, same as filter_size
        :param str padding: "same" or "valid"
        :param list[int]|int remove_padding:
        :param list[int|None]|int|None output_padding:
        :param bool with_bias: whether to add a bias. enabled by default
        """
        super().__init__(in_dim=in_dim, out_dim=out_dim, filter_size=filter_size, padding=padding, with_bias=with_bias)
        self.strides = strides
        self.remove_padding = remove_padding
        self.output_padding = output_padding

    def __call__(
        self,
        source: Tensor,
        *,
        in_spatial_dims: Sequence[Dim],
        out_spatial_dims: Optional[Sequence[Dim]] = None,
    ) -> Tuple[Tensor, Sequence[Dim]]:
        return transposed_conv(
            source,
            in_dim=self.in_dim,
            out_dim=self.out_dim,
            in_spatial_dims=in_spatial_dims,
            out_spatial_dims=out_spatial_dims,
            filter=self.filter,
            filter_size=self.filter_size,
            padding=self.padding,
            remove_padding=self.remove_padding,
            output_padding=self.output_padding,
            strides=self.strides,
            bias=self.bias if self.with_bias else None,
        )


# noinspection PyShadowingBuiltins
def transposed_conv(
    source: Tensor,
    *,
    in_dim: Dim,
    out_dim: Dim,
    in_spatial_dims: Sequence[Dim],
    out_spatial_dims: Optional[Sequence[Dim]] = None,
    filter: Tensor,
    filter_size: Sequence[Dim],
    padding: str,
    remove_padding: Union[Sequence[int], int] = 0,
    output_padding: Optional[Union[Sequence[Optional[int]], int]] = None,
    strides: Optional[Sequence[int]] = None,
    bias: Optional[Tensor] = None,
) -> Tuple[Tensor, Sequence[Dim]]:
    """transposed conv"""
    # noinspection PyProtectedMember
    out, out_spatial_dims = source._raw_backend.transposed_conv(
        source=source,
        in_dim=in_dim,
        out_dim=out_dim,
        in_spatial_dims=in_spatial_dims,
        out_spatial_dims=out_spatial_dims,
        filter=filter,
        filter_size=filter_size,
        padding=padding,
        remove_padding=remove_padding,
        output_padding=output_padding,
        strides=strides,
        bias=bias,
    )
    return out, out_spatial_dims


class TransposedConv1d(_TransposedConv):
    """
    1D transposed convolution
    """

    nd = 1

    __call__ = _ConvOrTransposedConv._call_nd1


class TransposedConv2d(_TransposedConv):
    """
    2D transposed convolution
    """

    nd = 2


class TransposedConv3d(_TransposedConv):
    """
    3D transposed convolution
    """

    nd = 3


def pool(
    source: Tensor,
    *,
    mode: str,
    pool_size: Union[Sequence[int], int],
    padding: str = "valid",
    dilation_rate: Union[Sequence[int], int] = 1,
    strides: Optional[Union[Sequence[int], int]] = None,
    in_spatial_dims: Union[Sequence[Dim], Dim],
    out_spatial_dims: Optional[Union[Sequence[Dim], Dim]] = None,
    nd: Optional[int] = None,
) -> Tuple[Tensor, Sequence[Dim]]:
    """
    A generic N-D pooling layer.
    This would usually be done after a convolution for down-sampling.

    :param Tensor source:
    :param nd:
    :param str mode: "max" or "avg"
    :param tuple[int] pool_size: shape of the window of each reduce
    :param str padding: "valid" or "same"
    :param tuple[int]|int dilation_rate:
    :param tuple[int]|int|None strides: in contrast to tf.nn.pool, the default (if it is None) will be set to pool_size
    :param Sequence[Dim] in_spatial_dims:
    :param Sequence[Dim]|None out_spatial_dims:
    :return: layer, out_spatial_dims
    """
    if isinstance(in_spatial_dims, Dim):
        in_spatial_dims = [in_spatial_dims]
    assert isinstance(in_spatial_dims, (list, tuple))
    assert all(isinstance(d, Dim) for d in in_spatial_dims)
    if nd is None:
        nd = len(in_spatial_dims)
    else:
        assert nd == len(in_spatial_dims)
    if out_spatial_dims is not None:
        if isinstance(out_spatial_dims, Dim):
            out_spatial_dims = [out_spatial_dims]
    if isinstance(pool_size, int):
        pool_size = [pool_size] * nd
    assert isinstance(pool_size, (list, tuple))
    assert len(pool_size) == nd
    if not strides:
        strides = pool_size
    elif isinstance(strides, int):
        strides = [strides] * nd
    assert isinstance(strides, (list, tuple))
    assert len(strides) == nd

    # noinspection PyProtectedMember
    out, out_spatial_dims = source._raw_backend.pool(
        source=source,
        mode=mode,
        pool_size=pool_size,
        padding=padding,
        dilation_rate=dilation_rate,
        strides=strides,
        in_spatial_dims=in_spatial_dims,
        out_spatial_dims=out_spatial_dims,
    )
    return out, out_spatial_dims


def max_pool(
    source: Tensor,
    *,
    pool_size: Union[Sequence[int], int],
    padding: str = "valid",
    dilation_rate: Union[Sequence[int], int] = 1,
    strides: Optional[Union[Sequence[int], int]] = None,
    in_spatial_dims: Union[Sequence[Dim], Dim],
    out_spatial_dims: Optional[Union[Sequence[Dim], Dim]] = None,
) -> Tuple[Tensor, Sequence[Dim]]:
    """max-pool"""
    return pool(
        source=source,
        mode="max",
        pool_size=pool_size,
        padding=padding,
        dilation_rate=dilation_rate,
        strides=strides,
        in_spatial_dims=in_spatial_dims,
        out_spatial_dims=out_spatial_dims,
    )


def max_pool1d(
    source: Tensor,
    *,
    pool_size: int,
    padding: str = "valid",
    dilation_rate: int = 1,
    strides: Optional[int] = None,
    in_spatial_dim: Dim,
    out_spatial_dim: Optional[Dim] = None,
) -> Tuple[Tensor, Dim]:
    """max pool"""
    return pool1d(
        source=source,
        mode="max",
        pool_size=pool_size,
        padding=padding,
        dilation_rate=dilation_rate,
        strides=strides,
        in_spatial_dim=in_spatial_dim,
        out_spatial_dim=out_spatial_dim,
    )


def pool1d(
    source: Tensor,
    *,
    mode: str,
    pool_size: int,
    padding: str = "valid",
    dilation_rate: int = 1,
    strides: Optional[int] = None,
    in_spatial_dim: Dim,
    out_spatial_dim: Optional[Dim] = None,
) -> Tuple[Tensor, Dim]:
    """
    1D pooling.

    :param Tensor source:
    :param str mode: "max" or "avg"
    :param tuple[int] pool_size: shape of the window of each reduce
    :param str padding: "valid" or "same"
    :param tuple[int]|int dilation_rate:
    :param tuple[int]|int|None strides: in contrast to tf.nn.pool, the default (if it is None) will be set to pool_size
    :param Sequence[Dim] in_spatial_dim:
    :param Sequence[Dim]|None out_spatial_dim:
    :return: layer, out_spatial_dim
    """
    assert isinstance(in_spatial_dim, Dim)
    out, (out_spatial_dim,) = pool(
        source=source,
        mode=mode,
        pool_size=pool_size,
        padding=padding,
        dilation_rate=dilation_rate,
        strides=strides,
        in_spatial_dims=[in_spatial_dim],
        out_spatial_dims=[out_spatial_dim] if out_spatial_dim is not None else None,
    )
    return out, out_spatial_dim


def pool2d(
    source: Tensor,
    *,
    mode: str,
    pool_size: Union[Sequence[int], int],
    padding: str = "valid",
    dilation_rate: Union[Sequence[int], int] = 1,
    strides: Optional[Union[Sequence[int], int]] = None,
    in_spatial_dims: Sequence[Dim],
    out_spatial_dims: Optional[Sequence[Dim]] = None,
) -> Tuple[Tensor, Sequence[Dim]]:
    """
    2D pooling.

    :param Tensor source:
    :param str mode: "max" or "avg"
    :param tuple[int] pool_size: shape of the window of each reduce
    :param str padding: "valid" or "same"
    :param tuple[int]|int dilation_rate:
    :param tuple[int]|int|None strides: in contrast to tf.nn.pool, the default (if it is None) will be set to pool_size
    :param Sequence[Dim] in_spatial_dims:
    :param Sequence[Dim]|None out_spatial_dims:
    :return: layer, out_spatial_dims
    """
    assert len(in_spatial_dims) == 2
    return pool(
        source=source,
        mode=mode,
        pool_size=pool_size,
        padding=padding,
        dilation_rate=dilation_rate,
        strides=strides,
        in_spatial_dims=in_spatial_dims,
        out_spatial_dims=out_spatial_dims,
    )


def pool3d(
    source: Tensor,
    *,
    mode: str,
    pool_size: Union[Sequence[int], int],
    padding: str = "valid",
    dilation_rate: Union[Sequence[int], int] = 1,
    strides: Optional[Union[Sequence[int], int]] = None,
    in_spatial_dims: Sequence[Dim],
    out_spatial_dims: Optional[Sequence[Dim]] = None,
) -> Tuple[Tensor, Sequence[Dim]]:
    """
    3D pooling.

    :param Tensor source:
    :param str mode: "max" or "avg"
    :param tuple[int] pool_size: shape of the window of each reduce
    :param str padding: "valid" or "same"
    :param tuple[int]|int dilation_rate:
    :param tuple[int]|int|None strides: in contrast to tf.nn.pool, the default (if it is None) will be set to pool_size
    :param Sequence[Dim] in_spatial_dims:
    :param Sequence[Dim]|None out_spatial_dims:
    :return: layer, out_spatial_dims
    """
    assert len(in_spatial_dims) == 3
    return pool(
        source=source,
        mode=mode,
        pool_size=pool_size,
        padding=padding,
        dilation_rate=dilation_rate,
        strides=strides,
        in_spatial_dims=in_spatial_dims,
        out_spatial_dims=out_spatial_dims,
    )


def make_conv_out_spatial_dims(
    in_spatial_dims: Sequence[Dim],
    *,
    filter_size: Union[Sequence[Union[int, Dim]], int, Dim],
    padding: str,
    strides: Union[Sequence[int], int] = 1,
    dilation_rate: Union[Sequence[int], int] = 1,
    description_prefix: Optional[str] = None,
) -> Sequence[Dim]:
    """create out spatial dims from in spatial dims"""
    nd = len(in_spatial_dims)
    if isinstance(filter_size, (int, Dim)):
        filter_size = [filter_size] * nd
    filter_size = [d.dimension if isinstance(d, Dim) else d for d in filter_size]
    assert all(isinstance(s, int) for s in filter_size)
    if isinstance(strides, int):
        strides = [strides] * nd
    if isinstance(dilation_rate, int):
        dilation_rate = [dilation_rate] * nd
    assert nd == len(in_spatial_dims) == len(filter_size) == len(strides) == len(dilation_rate)
    assert padding.lower() in ("valid", "same")
    out_spatial_dims = []
    for i in range(nd):
        in_spatial_dim = in_spatial_dims[i]
        if filter_size[i] == strides[i] == 1 or (strides[i] == 1 and padding.lower() == "same"):
            out_spatial_dims.append(in_spatial_dim)
        else:
            out_spatial_dim = _calc_out_dim(
                in_dim=in_spatial_dim,
                filter_size=filter_size[i],
                stride=strides[i],
                dilation_rate=dilation_rate[i],
                padding=padding,
            )
            assert isinstance(out_spatial_dim, Dim)
            if description_prefix and out_spatial_dim != in_spatial_dim:
                out_spatial_dim.name = f"{description_prefix}:spatial{i}"
            if in_spatial_dim.dyn_size_ext and not out_spatial_dim.dyn_size_ext:
                out_spatial_dim.dyn_size_ext = _calc_out_dim(
                    in_dim=in_spatial_dim.dyn_size_ext,
                    filter_size=filter_size[i],
                    stride=strides[i],
                    dilation_rate=dilation_rate[i],
                    padding=padding,
                )
            out_spatial_dims.append(out_spatial_dim)
    return out_spatial_dims


def _calc_out_dim(in_dim, filter_size, stride, padding, dilation_rate=1):
    """
    Copied and adapted from TF ConvLayer.calc_out_dim.

    :param T|int|Tensor|torch.Tensor|tensorflow.Tensor|Dim in_dim: dimension in some axis
    :param int filter_size: e.g. 2, for the corresponding axis
    :param int stride: e.g. 1, for the corresponding axis
    :param int dilation_rate: e.g. 1
    :param str padding: "valid" or "same"
    :return: the output dimension
    :rtype: T
    """

    def ceildiv(a, b):
        """
        :param T|int|Tensor|torch.Tensor|tensorflow.Tensor a:
        :param T|int|Tensor|torch.Tensor|tensorflow.Tensor b:
        :rtype: T
        """
        if isinstance(b, int) and b == 1:
            return a
        if isinstance(a, Tensor):
            return rf.ceil_divide(a, b)
        return -(-a // b)

    padding = padding.upper()
    # See tf.compat.v1.nn.convolution() documentation for more.
    if padding == "SAME":
        if isinstance(in_dim, Dim):
            return in_dim.ceildiv_right(stride)
        return ceildiv(in_dim, stride)
    elif padding == "VALID":
        if isinstance(in_dim, Dim):
            filter_left_dilated = (filter_size - 1) * dilation_rate // 2
            filter_right_dilated = (filter_size - 1) * dilation_rate - filter_left_dilated
            valid_part = in_dim.sub_left(filter_left_dilated).sub_right(filter_right_dilated)
            return valid_part.ceildiv_right(stride)
        return ceildiv(in_dim - (filter_size - 1) * dilation_rate, stride)
    else:
        raise Exception("invalid padding %r" % padding)



# start test like this:  nosetests-2.7  tests/test_TFUtil.py

from __future__ import annotations

import _setup_test_env  # noqa

import logging
import tensorflow as tf
import os
import sys
from returnn.tf.updater import *
from returnn.tf.layers.base import LayerBase, Loss
import returnn.tf.compat as tf_compat
from returnn.log import log
from numpy.testing import assert_almost_equal
import unittest
import numpy.testing
import contextlib
from returnn.util import better_exchook


@contextlib.contextmanager
def make_scope():
    with tf.Graph().as_default() as graph:
        with tf_compat.v1.Session(graph=graph) as session:
            yield session


class DummyLoss(Loss):
    need_target = False

    def get_value(self):
        assert self.layer and isinstance(self.layer, DummyLayer)
        return self.layer._get_loss_value()

    def get_error(self):
        return None


class DummyLayer(LayerBase):
    def __init__(self, initial_value=0.0, loss_value_factor=1.0, **kwargs):
        super(DummyLayer, self).__init__(**kwargs)
        self.loss_value_factor = loss_value_factor
        self.x = self.add_param(tf.Variable(initial_value))
        self.output.placeholder = self.x

    def _get_loss_value(self):
        return self.loss_value_factor * self.x

    @classmethod
    def get_losses(cls, name, network, output, loss=None, reduce_func=None, layer=None, **kwargs):
        assert not loss
        loss = DummyLoss(base_network=network)
        return super(DummyLayer, cls).get_losses(
            name=name, network=network, output=output, loss=loss, reduce_func=reduce_func, layer=layer, **kwargs
        )

    @classmethod
    def get_out_data_from_opts(cls, name, **kwargs):
        from returnn.tf.util.basic import Data

        return Data(name="%s_output" % name, batch_dim_axis=None, shape=(), dtype="float32")  # scalar


def test_Updater_GradientDescent():
    with make_scope() as session:
        from returnn.tf.network import TFNetwork, ExternData
        from returnn.config import Config

        config = Config()
        network = TFNetwork(extern_data=ExternData(), train_flag=True)
        network.add_layer(name="output", layer_class=DummyLayer, initial_value=5.0, loss_value_factor=3.0)
        network.initialize_params(session=session)

        updater = Updater(config=config, network=network)
        updater.set_learning_rate(1.0, session=session)
        updater.set_trainable_vars(network.get_trainable_params())
        updater.init_optimizer_vars(session=session)
        session.run(updater.get_optim_op())
        # One gradient descent step from 3.0 * x: gradient is 3, thus 5 - 3 = 2.
        assert_almost_equal(session.run(network.get_default_output_layer().output.placeholder), 2.0)


def test_Updater_CustomUpdate():
    with make_scope() as session:
        from returnn.tf.network import TFNetwork, ExternData
        from returnn.config import Config
        from returnn.tf.util.basic import CustomUpdate

        config = Config()
        network = TFNetwork(extern_data=ExternData(), train_flag=True)
        layer = network.add_layer(name="output", layer_class=DummyLayer, initial_value=4.0)
        assert isinstance(layer, DummyLayer)
        network.initialize_params(session=session)

        class CustomUpdateAdd13(CustomUpdate):
            def update_var(self, var):
                return tf_compat.v1.assign_add(var, 13.0)

        CustomUpdateAdd13().set_on_var(layer.x)

        updater = Updater(config=config, network=network)
        updater.set_learning_rate(1000.0, session=session)  # should be ignored
        updater.set_trainable_vars(network.get_trainable_params())
        updater.init_optimizer_vars(session=session)
        session.run(updater.get_optim_op())
        # Should have applied CustomUpdateAdd13.
        assert_almost_equal(session.run(network.get_default_output_layer().output.placeholder), 17.0)


def test_add_check_numerics_ops():
    with make_scope() as session:
        x = tf.constant(3.0, name="x")
        y = tf_compat.v1.log(x * 3, name="y")
        assert isinstance(y, tf.Tensor)
        assert_almost_equal(session.run(y), numpy.log(9.0))
        check = add_check_numerics_ops([y])
        session.run(check)
        z1 = tf_compat.v1.log(x - 3, name="z1")
        assert str(session.run(z1)) == "-inf"
        z2 = tf_compat.v1.log(x - 4, name="z2")
        assert str(session.run(z2)) == "nan"
        check1 = add_check_numerics_ops([z1])
        try:
            session.run(check1)
        except tf.errors.InvalidArgumentError as exc:
            print("Expected exception: %r" % exc)
        else:
            assert False, "should have raised an exception"
        check2 = add_check_numerics_ops([z2])
        try:
            session.run(check2)
        except tf.errors.InvalidArgumentError as exc:
            print("Expected exception: %r" % exc)
        else:
            assert False, "should have raised an exception"


def test_grad_add_check_numerics_ops():
    # Also see test_where_nan().
    with make_scope() as session:
        x = tf.Variable(initial_value=0.0, name="x")
        session.run(x.initializer)
        y = 1.0 / x
        grad_x = tf.gradients(y, x)[0]
        print("grad_x:", grad_x.eval())
        assert str(float("-inf")) == "-inf"
        assert str(grad_x.eval()) == "-inf"

        session.run(x.assign(1.0))
        opt = tf_compat.v1.train.GradientDescentOptimizer(learning_rate=1.0)
        train_op = opt.minimize(y, var_list=[x])
        check = add_check_numerics_ops([train_op])
        session.run(check)

        session.run(x.assign(0.0))
        try:
            session.run(check)  # should fail now
        except tf.errors.InvalidArgumentError as exc:
            print("Expected exception: %r" % exc)
        else:
            assert False, "should have raised an exception"


def test_Updater_add_check_numerics_ops():
    class _Layer(DummyLayer):
        def _get_loss_value(self):
            return tf_compat.v1.log(self.x)

    from returnn.tf.network import TFNetwork, ExternData
    from returnn.config import Config

    with make_scope() as session:
        config = Config()
        config.set("debug_add_check_numerics_ops", True)
        network = TFNetwork(extern_data=ExternData(), train_flag=True)
        network.add_layer(name="output", layer_class=_Layer, initial_value=1.0)
        network.initialize_params(session=session)

        updater = Updater(config=config, network=network)
        updater.set_learning_rate(1.0, session=session)
        updater.set_trainable_vars(network.get_trainable_params())
        updater.init_optimizer_vars(session=session)
        # Should succeed.
        session.run(updater.get_optim_op())
        # One gradient descent step from ln(x), x = 1.0: gradient is 1.0 / x, thus x - 1.0 = 0.0.
        assert_almost_equal(session.run(network.get_default_output_layer().output.placeholder), 0.0)

        try:
            # Now, should fail.
            session.run(updater.get_optim_op())
        except tf.errors.InvalidArgumentError as exc:
            print("Expected exception: %r" % exc)
        else:
            assert False, "should have raised an exception"


def test_Updater_simple_batch():
    with make_scope() as session:
        from returnn.tf.network import TFNetwork, ExternData
        from returnn.config import Config
        from returnn.datasets.generating import Task12AXDataset

        dataset = Task12AXDataset()
        dataset.init_seq_order(epoch=1)
        extern_data = ExternData()
        extern_data.init_from_dataset(dataset)

        config = Config()
        network = TFNetwork(extern_data=extern_data, train_flag=True)
        network.construct_from_dict(
            {
                "layer1": {"class": "linear", "activation": "tanh", "n_out": 13, "from": "data:data"},
                "layer2": {"class": "linear", "activation": "tanh", "n_out": 13, "from": ["layer1"]},
                "output": {"class": "softmax", "loss": "ce", "target": "classes", "from": ["layer2"]},
            }
        )
        network.initialize_params(session=session)

        updater = Updater(config=config, network=network)
        updater.set_learning_rate(1.0, session=session)
        updater.set_trainable_vars(network.get_trainable_params())
        updater.init_optimizer_vars(session=session)

        from returnn.tf.data_pipeline import FeedDictDataProvider

        batches = dataset.generate_batches(
            recurrent_net=network.recurrent,
            batch_size=100,
            max_seqs=10,
            max_seq_length=sys.maxsize,
            used_data_keys=network.used_data_keys,
        )
        data_provider = FeedDictDataProvider(
            extern_data=extern_data,
            data_keys=network.used_data_keys,
            dataset=dataset,
            batches=batches,
        )
        feed_dict, _ = data_provider.get_feed_dict(single_threaded=True)
        session.run(updater.get_optim_op(), feed_dict=feed_dict)


def test_Updater_decouple_constraints():
    with make_scope() as session:
        from returnn.tf.network import TFNetwork, ExternData
        from returnn.config import Config
        from returnn.datasets.generating import Task12AXDataset

        dataset = Task12AXDataset()
        dataset.init_seq_order(epoch=1)
        extern_data = ExternData()
        extern_data.init_from_dataset(dataset)

        config = Config({"decouple_constraints": True})
        network = TFNetwork(extern_data=extern_data, train_flag=True)
        network.construct_from_dict(
            {
                "layer1": {"class": "linear", "activation": "tanh", "n_out": 13, "L2": 0.01, "from": "data:data"},
                "layer2": {"class": "linear", "activation": "tanh", "n_out": 13, "L2": 0.01, "from": "layer1"},
                "output": {"class": "softmax", "loss": "ce", "target": "classes", "from": "layer2"},
            }
        )
        network.initialize_params(session=session)

        updater = Updater(config=config, network=network)
        updater.set_learning_rate(1.0, session=session)
        updater.set_trainable_vars(network.get_trainable_params())
        updater.init_optimizer_vars(session=session)
        update_op = updater.get_optim_op()
        assert updater.decouple_constraints

        from returnn.tf.data_pipeline import FeedDictDataProvider

        batches = dataset.generate_batches(
            recurrent_net=network.recurrent,
            batch_size=100,
            max_seqs=10,
            max_seq_length=sys.maxsize,
            used_data_keys=network.used_data_keys,
        )
        data_provider = FeedDictDataProvider(
            extern_data=extern_data,
            data_keys=network.used_data_keys,
            dataset=dataset,
            batches=batches,
        )
        feed_dict, _ = data_provider.get_feed_dict(single_threaded=True)
        session.run(update_op, feed_dict=feed_dict)


def test_Updater_decouple_constraints_simple_graph():
    with make_scope() as session:
        from returnn.tf.network import TFNetwork, ExternData
        from returnn.config import Config

        config = Config(
            {
                "decouple_constraints": True,
                "decouple_constraints_factor": 1.0,
            }
        )
        extern_data = ExternData({"data": dict(shape=(), dtype="float32")})
        network = TFNetwork(config=config, extern_data=extern_data, train_flag=True)
        network.construct_from_dict(
            {
                "var": {"class": "variable", "shape": (), "init": 1.0, "L2": 0.01},
                "loss": {
                    "class": "eval",
                    "from": ["data", "var"],
                    "eval": "source(0, auto_convert=False) * source(1, auto_convert=False)",
                    "loss": "as_is",
                },
            }
        )
        network.initialize_params(session=session)
        var = network.get_layer("var").output.placeholder
        assert session.run(var) == 1.0

        updater = Updater(config=config, network=network, initial_learning_rate=1.0)
        updater.set_learning_rate(1.0, session=session)
        updater.set_trainable_vars(network.get_trainable_params())
        updater.init_optimizer_vars(session=session)
        update_op = updater.get_optim_op()
        assert updater.decouple_constraints

        tf_util.print_graph_output(update_op)

        session.run(
            update_op,
            feed_dict={extern_data.data["data"].placeholder: [0.0, 0.0, 0.0], extern_data.get_batch_info().dim: 3},
        )
        assert_almost_equal(session.run(var), 1.0 - 2.0 * 0.01)


def test_Updater_decouple_constraints_simple_graph_grad_accum():
    with make_scope() as session:
        from returnn.tf.network import TFNetwork, ExternData
        from returnn.config import Config

        config = Config(
            {
                "decouple_constraints": True,
                "decouple_constraints_factor": 1.0,
                "accum_grad_multiple_step": 2,
            }
        )
        extern_data = ExternData({"data": dict(shape=(), dtype="float32")})
        network = TFNetwork(config=config, extern_data=extern_data, train_flag=True)
        network.construct_from_dict(
            {
                "var": {"class": "variable", "shape": (), "init": 1.0, "L2": 0.01},
                "loss": {
                    "class": "eval",
                    "from": ["data", "var"],
                    "eval": "source(0, auto_convert=False) * source(1, auto_convert=False)",
                    "loss": "as_is",
                },
            }
        )
        network.initialize_params(session=session)
        var = network.get_layer("var").output.placeholder
        assert session.run(var) == 1.0

        updater = Updater(config=config, network=network, initial_learning_rate=1.0)
        updater.set_learning_rate(1.0, session=session)
        updater.set_trainable_vars(network.get_trainable_params())
        updater.init_optimizer_vars(session=session)
        update_op = updater.get_optim_op()
        assert updater.decouple_constraints

        tf_util.print_graph_output(update_op)

        assert session.run(updater.global_train_step) == 0
        expected_var = 1.0
        for i in range(10):
            print("Run step", i)
            session.run(
                update_op,
                feed_dict={extern_data.data["data"].placeholder: [0.0, 0.0, 0.0], extern_data.get_batch_info().dim: 3},
            )
            assert session.run(updater.global_train_step) == i + 1
            var_value = session.run(var)
            print("var:", var_value)
            if i % 2 == 1:
                expected_var -= 2.0 * 0.01 * expected_var
            assert_almost_equal(var_value, expected_var)


def test_Updater_multiple_optimizers():
    with make_scope() as session:
        from returnn.tf.network import TFNetwork, ExternData
        from returnn.config import Config
        from returnn.datasets.generating import Task12AXDataset

        dataset = Task12AXDataset()
        dataset.init_seq_order(epoch=1)
        extern_data = ExternData()
        extern_data.init_from_dataset(dataset)

        config = Config()
        network = TFNetwork(extern_data=extern_data, train_flag=True)
        network.construct_from_dict(
            {
                "layer1": {
                    "class": "linear",
                    "activation": "tanh",
                    "n_out": 13,
                    "from": "data:data",
                    "updater_opts": {"optimizer": {"class": "Adam"}},
                },
                "layer2": {
                    "class": "linear",
                    "activation": "tanh",
                    "n_out": 13,
                    "from": ["layer1"],
                    "updater_opts": {"optimizer": {"class": "Adagrad"}},
                },
                "output": {"class": "softmax", "loss": "ce", "target": "classes", "from": ["layer2"]},
            }
        )
        network.initialize_params(session=session)

        updater = Updater(config=config, network=network)
        updater.set_learning_rate(1.0, session=session)
        updater.set_trainable_vars(network.get_trainable_params())
        updater.init_optimizer_vars(session=session)

        optim_op = updater.get_optim_op()
        assert len(updater.optimizers) == 3

        from returnn.tf.data_pipeline import FeedDictDataProvider

        batches = dataset.generate_batches(
            recurrent_net=network.recurrent,
            batch_size=100,
            max_seqs=10,
            max_seq_length=sys.maxsize,
            used_data_keys=network.used_data_keys,
        )
        data_provider = FeedDictDataProvider(
            extern_data=extern_data,
            data_keys=network.used_data_keys,
            dataset=dataset,
            batches=batches,
        )
        feed_dict, _ = data_provider.get_feed_dict(single_threaded=True)
        session.run(optim_op, feed_dict=feed_dict)


def test_Updater_multiple_optimizers_and_opts():
    with make_scope() as session:
        from returnn.tf.network import TFNetwork, ExternData
        from returnn.config import Config
        from returnn.datasets.generating import Task12AXDataset

        dataset = Task12AXDataset()
        dataset.init_seq_order(epoch=1)
        extern_data = ExternData()
        extern_data.init_from_dataset(dataset)

        config = Config()
        network = TFNetwork(extern_data=extern_data, train_flag=True)
        network.construct_from_dict(
            {
                "layer1": {
                    "class": "linear",
                    "activation": "tanh",
                    "n_out": 13,
                    "from": "data:data",
                    "updater_opts": {"optimizer": {"class": "Adam"}, "accum_grad_multiple_step": 2},
                },
                "layer2": {
                    "class": "linear",
                    "activation": "tanh",
                    "n_out": 13,
                    "from": ["layer1"],
                    "updater_opts": {
                        "optimizer": {"class": "Adagrad", "learning_rate_multiplier": 3},
                        "gradient_noise": 0.1,
                    },
                },
                "output": {"class": "softmax", "loss": "ce", "target": "classes", "from": ["layer2"]},
            }
        )
        network.initialize_params(session=session)

        updater = Updater(config=config, network=network)
        updater.set_learning_rate(1.0, session=session)
        updater.set_trainable_vars(network.get_trainable_params())
        updater.init_optimizer_vars(session=session)

        optim_op = updater.get_optim_op()
        assert len(updater.optimizers) == 3

        from returnn.tf.data_pipeline import FeedDictDataProvider

        batches = dataset.generate_batches(
            recurrent_net=network.recurrent,
            batch_size=100,
            max_seqs=10,
            max_seq_length=sys.maxsize,
            used_data_keys=network.used_data_keys,
        )
        data_provider = FeedDictDataProvider(
            extern_data=extern_data,
            data_keys=network.used_data_keys,
            dataset=dataset,
            batches=batches,
        )
        feed_dict, _ = data_provider.get_feed_dict(single_threaded=True)
        session.run(optim_op, feed_dict=feed_dict)


if __name__ == "__main__":
    better_exchook.install()
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                try:
                    v()
                except unittest.SkipTest as exc:
                    print("SkipTest:", exc)
                print("-" * 40)
        print("Finished all tests.")
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute



import sys
import os

import _setup_test_env  # noqa
from returnn.config import Config
from returnn.learning_rate_control import *
import numpy
import unittest

from returnn.util import better_exchook
from returnn.log import log


def test_save_load():
    import tempfile

    with tempfile.NamedTemporaryFile(mode="w") as f:
        filename = f.name
    assert not os.path.exists(filename)
    control = LearningRateControl(default_learning_rate=1.0, filename=filename)
    assert 2 not in control.epoch_data
    control.epoch_data[2] = LearningRateControl.EpochData(
        learning_rate=0.0008,
        error={
            "dev_error_ctc": 0.22486433815293946,
            "dev_error_decision": 0.0,
            "dev_error_output/output_prob": 0.16270349413262444,
            "dev_score_ctc": 1.0732941136466485,
            "dev_score_output/output_prob": 0.7378438060027533,
            "train_error_ctc": 0.13954045252681482,
            "train_error_decision": 0.0,
            "train_error_output/output_prob": 0.106904268810835,
            "train_score_ctc": 0.5132869609859635,
            "train_score_output/output_prob": 0.5098970897590558,
        },
    )
    control.save()
    assert os.path.exists(filename)
    control = LearningRateControl(default_learning_rate=1.0, filename=filename)
    assert 2 in control.epoch_data
    data = control.epoch_data[2]
    numpy.testing.assert_allclose(data.learning_rate, 0.0008)
    assert "dev_error_output/output_prob" in data.error
    numpy.testing.assert_allclose(data.error["dev_error_output/output_prob"], 0.16270349413262444)


def test_load():
    import tempfile

    with tempfile.NamedTemporaryFile(mode="w") as f:
        f.write(
            """{
      1: EpochData(learningRate=0.0008, error={
      'dev_error_ctc': 0.21992561489090365,
      'dev_error_decision': 0.0,
      'dev_error_output/output_prob': 0.1597158714185534,
      'dev_score_ctc': 1.0742086989480388,
      'dev_score_output/output_prob': 0.7316125233255415,
      'train_error_ctc': 0.11740542462939381,
      'train_error_decision': 0.0,
      'train_error_output/output_prob': 0.10000902651529825,
      'train_score_ctc': 0.42154947919396146,
      'train_score_output/output_prob': 0.4958179737218142,
      }),
      2: EpochData(learningRate=0.0008, error={
      'dev_error_ctc': 0.22486433815293946,
      'dev_error_decision': 0.0,
      'dev_error_output/output_prob': 0.16270349413262444,
      'dev_score_ctc': 1.0732941136466485,
      'dev_score_output/output_prob': 0.7378438060027533,
      'train_error_ctc': 0.13954045252681482,
      'train_error_decision': 0.0,
      'train_error_output/output_prob': 0.106904268810835,
      'train_score_ctc': 0.5132869609859635,
      'train_score_output/output_prob': 0.5098970897590558,
      }),
    }"""
        )
        f.flush()
        control = LearningRateControl(default_learning_rate=1.0, filename=f.name)
        assert set(control.epoch_data.keys()) == {1, 2}
        data = control.epoch_data[2]
        numpy.testing.assert_allclose(data.learning_rate, 0.0008)
        assert "dev_error_output/output_prob" in data.error
        numpy.testing.assert_allclose(data.error["dev_error_output/output_prob"], 0.16270349413262444)


def test_init_error_old():
    config = Config()
    config.update({"learning_rate_control": "newbob", "learning_rate_control_error_measure": "dev_score"})
    lrc = load_learning_rate_control_from_config(config)
    assert isinstance(lrc, NewbobRelative)
    lrc.get_learning_rate_for_epoch(1)
    lrc.set_epoch_error(1, {"train_score": 1.9344199658230012})
    lrc.set_epoch_error(1, {"dev_score": 1.99, "dev_error": 0.6})
    error = lrc.get_epoch_error_dict(1)
    assert "train_score" in error
    assert "dev_score" in error
    assert "dev_error" in error
    assert lrc.get_error_key(1) == "dev_score"
    lrc.get_learning_rate_for_epoch(2)
    lrc.set_epoch_error(2, {"train_score": 1.8})
    lrc.set_epoch_error(2, {"dev_score": 1.9, "dev_error": 0.5})
    lrc.get_learning_rate_for_epoch(3)


def test_init_error_new():
    config = Config()
    config.update({"learning_rate_control": "newbob", "learning_rate_control_error_measure": "dev_score"})
    lrc = load_learning_rate_control_from_config(config)
    assert isinstance(lrc, NewbobRelative)
    lrc.get_learning_rate_for_epoch(1)
    lrc.set_epoch_error(1, {"train_score": {"cost:output": 1.9344199658230012}})
    lrc.set_epoch_error(1, {"dev_score": {"cost:output": 1.99}, "dev_error": {"error:output": 0.6}})
    error = lrc.get_epoch_error_dict(1)
    assert "train_score" in error
    assert "dev_score" in error
    assert "dev_error" in error
    assert lrc.get_error_key(1) == "dev_score"
    lrc.get_learning_rate_for_epoch(2)
    lrc.set_epoch_error(2, {"train_score": {"cost:output": 1.8}})
    lrc.set_epoch_error(2, {"dev_score": {"cost:output": 1.9}, "dev_error": {"error:output": 0.5}})
    lrc.get_learning_rate_for_epoch(3)


def test_init_error_muliple_out():
    config = Config()
    config.update({"learning_rate_control": "newbob", "learning_rate_control_error_measure": "dev_score"})
    lrc = load_learning_rate_control_from_config(config)
    assert isinstance(lrc, NewbobRelative)
    lrc.get_learning_rate_for_epoch(1)
    lrc.set_epoch_error(1, {"train_score": {"cost:output": 1.95, "cost:out2": 2.95}})
    lrc.set_epoch_error(
        1,
        {"dev_score": {"cost:output": 1.99, "cost:out2": 2.99}, "dev_error": {"error:output": 0.6, "error:out2": 0.7}},
    )
    error = lrc.get_epoch_error_dict(1)
    assert "train_score_output" in error
    assert "train_score_out2" in error
    assert "dev_score_output" in error
    assert "dev_score_out2" in error
    assert "dev_error_output" in error
    assert "dev_error_out2" in error
    assert lrc.get_error_key(1) == "dev_score_output"
    lrc.get_learning_rate_for_epoch(2)
    lrc.set_epoch_error(2, {"train_score": {"cost:output": 1.8, "cost:out2": 2.8}})
    lrc.set_epoch_error(
        2, {"dev_score": {"cost:output": 1.9, "cost:out2": 2.9}, "dev_error": {"error:output": 0.5, "error:out2": 0.6}}
    )
    lrc.get_learning_rate_for_epoch(3)


def test_newbob():
    lr = 0.01
    config = Config()
    config.update({"learning_rate_control": "newbob", "learning_rate": lr})
    lrc = load_learning_rate_control_from_config(config)
    assert isinstance(lrc, NewbobRelative)
    assert lrc.get_learning_rate_for_epoch(1) == lr
    lrc.set_epoch_error(1, {"train_score": {"cost:output": 1.9344199658230012}})
    lrc.set_epoch_error(1, {"dev_score": {"cost:output": 1.99}, "dev_error": {"error:output": 0.6}})
    error = lrc.get_epoch_error_dict(1)
    assert "train_score" in error
    assert "dev_score" in error
    assert "dev_error" in error
    assert lrc.get_error_key(1) == "dev_score"
    assert lrc.get_learning_rate_for_epoch(2) == lr  # epoch 2 cannot be a different lr yet
    lrc.set_epoch_error(2, {"train_score": {"cost:output": 1.8}})
    lrc.set_epoch_error(2, {"dev_score": {"cost:output": 1.9}, "dev_error": {"error:output": 0.5}})
    lrc.get_learning_rate_for_epoch(3)


def test_newbob_multi_epoch():
    lr = 0.0005
    config = Config()
    config.update(
        {
            "learning_rate_control": "newbob_multi_epoch",
            "learning_rate_control_relative_error_relative_lr": True,
            "newbob_multi_num_epochs": 6,
            "newbob_multi_update_interval": 1,
            "learning_rate": lr,
        }
    )
    lrc = load_learning_rate_control_from_config(config)
    assert isinstance(lrc, NewbobMultiEpoch)
    assert lrc.get_learning_rate_for_epoch(1) == lr
    lrc.set_epoch_error(
        1,
        {
            "dev_error": 0.50283176046904721,
            "dev_score": 2.3209858321263455,
            "train_score": 3.095824052426714,
        },
    )
    assert lrc.get_learning_rate_for_epoch(2) == lr  # epoch 2 cannot be a different lr yet


def test_later_default_lr():
    import tempfile

    tmp_file = tempfile.mktemp()

    lr = 0.0005
    learning_rates = list(numpy.linspace(0.0003, lr, num=10))  # warmup
    config = Config()
    config.update(
        {
            "learning_rate_file": tmp_file,
            "learning_rate_control": "newbob_multi_epoch",
            "learning_rate_control_relative_error_relative_lr": True,
            "learning_rate_control_min_num_epochs_per_new_lr": 3,
            "newbob_multi_num_epochs": 6,
            "newbob_multi_update_interval": 1,
            "learning_rate": lr,
            "learning_rates": learning_rates,
            "min_learning_rate": lr / 50.0,
        }
    )
    lrc = load_learning_rate_control_from_config(config)
    assert isinstance(lrc, NewbobMultiEpoch)

    num_epochs = 250
    for epoch in range(1, num_epochs + 1):
        lrc.get_learning_rate_for_epoch(epoch)
        lrc.set_epoch_error(epoch, {"train_score": 0.5, "train_error": 0.5})
        lrc.set_epoch_error(epoch, {"dev_score": 0.5, "dev_error": 0.5})
    print("Learning rates:")
    print(lrc)
    lrc.save()
    print("Saved to:", lrc.filename)

    learning_rates = {i + 1: v for (i, v) in enumerate(learning_rates)}
    later_epoch = num_epochs + 1
    learning_rates[later_epoch] = lr * 0.5  # reset
    config.update({"learning_rates": learning_rates})

    lrc = load_learning_rate_control_from_config(config)
    assert later_epoch in lrc.epoch_data
    lr251 = lrc.get_learning_rate_for_epoch(later_epoch)
    numpy.testing.assert_almost_equal(lr251, learning_rates[later_epoch])


if __name__ == "__main__":
    better_exchook.install()
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                try:
                    v()
                except unittest.SkipTest as exc:
                    print("SkipTest:", exc)
                print("-" * 40)
        print("Finished all tests.")
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute
